# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https: // mozilla.org / MPL / 2.0 / .
# Copyright 2021 Peter Dunne



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https: // mozilla.org / MPL / 2.0 / .
# Copyright 2021 Peter Dunne
"""Imports all magnet classes and routines
"""
from . import plot
from . import magnets
from .magnets._magnet import reset_magnets, list_magnets
from .magnets._routines2 import grid2D, B_calc_2D
from .magnets._routines3 import grid3D, B_calc_3D
from math import pi as PI

PI_2 = PI / 2.0
PI_4 = PI / 4.0

u0 = 4E-7 * PI

__all__ = ['magnets', 'plot']



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https: // mozilla.org / MPL / 2.0 / .
# Copyright 2021 Peter Dunne
"""Magnet Base Class
"""
# __all__ = ['Magnet', 'reset_magnets', 'list_magnets']
__all__ = ['Magnet']

from typing import List, Any, Union, Type, Optional
from weakref import WeakSet


class Registry:
    """Registry class for tracking instances

    Instances are tracked in `class.instances` using weak references.
    This also includes any instances that are deleted manually or go out of
    scope. 

    Class methods:

        `print_instances()` 

        `get_instances()`
        
        `get_num_instances()`
    """    
    instances = WeakSet()
    _hard_instances = []

    def __new__(cls, *args, **kwargs):
        o = object.__new__(cls)
        cls._register_instance(o)
        return o

    def __init__(self, *args, **kwargs):
        super().__init__()
        
    @classmethod
    def print_instances(cls):
        """Prints class instantiations
        """        
        if len(cls.instances) < 1:
            print("No Instances")
        else:    
            for instance in cls.instances:
                print(instance)

    @classmethod
    def get_instances(cls):
        return(cls.instances)
    
    @classmethod
    def get_num_instances(cls, Print_Val=False):
        """Return number of instances of class

        Args:
            Print_Val (bool, optional): [Print to screen]. Defaults to False.

        Returns:
            num_instances [int]: 
        """        
        if Print_Val:
            print(len(cls.instances))
        return len(cls.instances)

    @classmethod
    def _register_instance(cls, instance):
        """[summary]

        Args:
            instance ([type]): [description]
        """        
        cls.instances.add(instance)
        cls._hard_instances.append(instance)
        for b in cls.__bases__:
            if issubclass(b, Registry):
                b._register_instance(instance)
                
    @classmethod
    def reset(cls) -> None:
        """Removes all magnet instances from registry.
        """
        for magnet in cls._hard_instances:
            del magnet
        cls.instances = WeakSet()
        cls._hard_instances = []

    def __init_subclass__(cls):
        cls.instances = WeakSet()
        cls._hard_instances = []


class Magnet(Registry):
    """Magnet base class


    Returns:
        Magnet: magnet base class
    """    
    counter = 0
    mag_type = 'Magnet'

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()


def reset_magnets():
    """Returns a list of all instantiated magnets. 
    """
    from ._magnet2 import Magnet_2D, Rectangle, Square
    from ._magnet3 import Magnet_3D, Prism, Cube, Cylinder
    magnet_classes = [Registry, Magnet, 
                        Magnet_2D, Rectangle, Square,
                        Magnet_3D, Prism, Cube, Cylinder]
    for cls in magnet_classes:
        cls.reset()

def list_magnets():
    """Returns a list of all instantiated magnets. 

    Assumes that the child class registries have not been modified outside of 
    using `pymagnet.reset_magnets()`.
    """    
    return Magnet.print_instances()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https: // mozilla.org / MPL / 2.0 / .
# Copyright 2021 Peter Dunne

import pkg_resources
from pathlib import Path
import unittest
import os
import sys
import pymagnet as pm  # Magnetic library
import numpy as np  # we use numpy for handing vectors and matrices
import matplotlib.pyplot as plt  # Matplotlib is used to generate our plots


_REQUIREMENTS_PATH = ('tests/requirements.txt')

class MagnetTest(unittest.TestCase):

    def test_python_major_version(self):
        """Requires at least Python 3.6"""
        err_msg = "Requires at least Python 3.6\nYou are using Python {}.{}.".format(
            sys.version_info.major, sys.version_info.minor)
        self.assertEqual(sys.version_info.major == 3 and sys.version_info.minor >= 6,
                         True, msg=err_msg)

    def test_requirements(self):
        """Test that each required package is available."""
        # Ref: https://stackoverflow.com/a/45474387/
        with open(_REQUIREMENTS_PATH) as f:
            requirements = pkg_resources.parse_requirements(f)
            for requirement in requirements:
                requirement = str(requirement)
                with self.subTest(requirement=requirement):
                    pkg_resources.require(requirement)

    def test_create_cylinder(self):
        """Creating a cylinder magnet"""
        R, L, Jr = 1.0, 1.0, 1.0
        center = (0.0, 0.0, 0.0)
        self.assertEqual(pm.magnets.Cylinder(R, L, Jr, center=center).mag_type,
                         'Cylinder')

    def test_create_prism(self):
        """Creating a cuboid magnet"""
        a, b, c, Jr = 1.0, 1.0, 1.0, 1.0
        center = (0.0, 0.0, 0.0)
        self.assertEqual(pm.magnets.Prism(a, b, c, Jr, center=center).mag_type,
                         'Prism')

    def test_create_cube(self):
        """Creating a cube magnet"""
        a, Jr = 1.0, 1.0
        center = (0.0, 0.0, 0.0)
        self.assertEqual(pm.magnets.Cube(a, Jr, center=center).mag_type,
                         'Cube')

    def test_create_rectangle(self):
        """Creating a 2D rectangular magnet"""
        a, b, Jr = 1.0, 1.0, 1.0
        center = (0.0, 0.0)
        self.assertEqual(pm.magnets.Rectangle(a, b, Jr, center=center).mag_type,
                         'Rectangle')

    def test_create_square(self):
        """Creating a 2D square magnet"""
        a, Jr = 1.0, 1.0,
        center = (0.0, 0.0)
        self.assertEqual(pm.magnets.Square(a, Jr, center=center).mag_type,
                         'Square')

    def test_cylinder_class_instance(self):
        """Creating another Cylinder"""
        R, L, Jr = 1.0, 1.0, 1.0
        center = (0.0, 0.0, 0.0)
        magnet = pm.magnets.Cylinder(radius=R, length=L, Jr=Jr, center=center)
        self.assertIsInstance(magnet, pm.magnets.Cylinder)

    def test_1D_cylinder_surface_max(self):
        """Cylinder: Central magnetic field on surface"""
        R, L = 1.0, 2.0
        Jr = 1.0
        zc = -L/2
        m_cyl = pm.magnets.Cylinder(
            radius=R, length=L, Jr=Jr, center=(0.0, 0.0, zc))
        Bz = pm.magnets.magnetic_field_cylinder_1D(m_cyl, 0.0)
        self.assertEqual(round(Bz, 5), 0.44721)

    def test_1D_cylinder_center_value(self):
        """Cylinder: Magnetic field at magnet center"""
        R, L = 1.0, 2.0
        Jr = 1.0
        zc = -L/2
        m_cyl = pm.magnets.Cylinder(
            radius=R, length=L, Jr=Jr, center=(0.0, 0.0, zc))

        Bz = pm.magnets.magnetic_field_cylinder_1D(m_cyl, zc)
        self.assertEqual(round(Bz, 5), 0.70711)

    def test_1D_prism_surface_max(self):
        """Cuboid: Central magnetic field on surface"""
        width = 1.0
        depth = 2.0
        height = 3.0
        Jr = 1.0
        zc = -height / 2
        m_quad = pm.magnets.Prism(width=width, depth=depth, height=height,
                                    Jr=Jr, center=(0.0, 0.0, zc))
        Bz = pm.magnets.magnetic_field_prism_1D(m_quad, 0)
        self.assertEqual(round(Bz, 5), 0.48344)

    def test_1D_prism_center_value(self):
        """Cuboid: Magnetic field at magnet center\n"""
        width = 1.0
        depth = 2.0
        height = 3.0
        Jr = 1.0
        zc = -height / 2
        m_quad = pm.magnets.Prism(width=width, depth=depth, height=height,
                                   Jr=Jr, center=(0.0, 0.0, zc))
        Bz = pm.magnets.magnetic_field_prism_1D(m_quad, zc)
        self.assertEqual(round(Bz, 5), 0.88775)

    def test_2D_two_magnets(self):
        """Test 2 Rectangular Magnets\n"""

        pm.reset_magnets()
        width = 20e-3
        height = 40e-3
        center = (-0.75*width, 0)
        mag1 = pm.magnets.Rectangle(
            width=width, height=height, Jr=1.0, center=center, theta=45.0)
        center = (0.75*width, 0)
        mag2 = pm.magnets.Rectangle(
            width=width, height=height, Jr=1.0, center=center, theta=-45.0)

        x, y = 0.0, height/2
        B = pm.magnets.B_calc_2D(x, y)
        result = round(B.n[0], 5)
        self.assertEqual(result, 0.42462)

    def test_3D_four_cubes(self):
        """Test 4 Cubes in Pseudo-Quadrupolar Arrangement"""

        pm.reset_magnets()
        width = 10e-3
        a = width/2
        hGap = a
        theta, phi = 0.0, 90.0

        # Add top left magnet
        _ = pm.magnets.Cube(width=width, Jr=1.0, center=(-a - hGap,
                                                 0, a), theta=theta, phi=phi)
        # Add bottom left magnet
        _ = pm.magnets.Cube(
            a=a, Jr=-1.0, center=(-a - hGap, 0, -a), theta=theta, phi=phi)
        # Add top right magnet
        _ = pm.magnets.Cube(a=a, Jr=1.0, center=(
            a + hGap, 0, a), theta=theta, phi=phi)
        # Add bottom right magnet
        _ = pm.magnets.Cube(a=a, Jr=-1.0, center=(a + hGap,
                                                  0, -a), theta=theta, phi=phi)

        x, y, z = a/2, a, a/3
        result = round(pm.magnets.B_calc_3D(x, y, z).n[0], 5)
        self.assertEqual(result, 0.15011)
    
    def test_solenoid_off_axis(self):
        """Off-axis test of solenoid magnetic field"""

        R = 5e-3
        L = 20e-3
        m_cyl = pm.magnets.Cylinder(radius=R, length=L, Jr=1.0,
                            center=(0.0, 0.0, 0))

        result = round(np.linalg.norm(
            m_cyl._calcB_cyl(m_cyl.radius * .4, m_cyl.length * .2)), 5)
        self.assertEqual(result, 0.86351)


if __name__ == "__main__":
    unittest.main(argv=[''], verbosity=2,)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https: // mozilla.org / MPL / 2.0 / .
# Copyright 2021 Peter Dunne
"""Three Dimensional Magnet Classes
"""
import numpy as _np
from ._fields import Point3
from ._magnet import Magnet
# from pymagma import PI

__all__ = ['Magnet_3D', 'Prism', 'Cube', 'Cylinder']

class Magnet_3D(Magnet):
    """3D Magnet Class

    Args:
         width: magnet width [default 10e-3]
         depth: magnet depth [default 20e-3] 
         height: magnet height [default 30e-3] 
         
    Optional Arguments:
         centre: magnet centre (Point3_3D object) [default Point3(.0, .0, .0)]
         J: remnant magnetisation in Tesla [default 1.0]

    """
    mag_type = 'Magnet_3D'

    def __init__(self, width, depth, height, Jr, **kwargs) -> None:
        super().__init__()
        self.width = width
        self.depth = depth
        self.height = height

        self.a = width/2
        self.b = depth/2
        self.c = height/2

        self.Jr = Jr 

        center = kwargs.pop('center', Point3(0.0, 0.0, 0.0))

        if type(center) is tuple:
            center = Point3(center[0], center[1], center[2])

        self.xc = center.x
        self.yc = center.y
        self.zc = center.z

    def center(self):
        return _np.array([self.xc, self.yc, self.zc])
    
    def size(self):
        """Returns magnet dimesions

        Returns:
            size[ndarray]: numpy array [width, depth, height]
        """
        return _np.array([self.width, self.depth, self.height])



class Prism(Magnet_3D):
    """Prism Magnet Class. Cuboid width x depth x height.

   Args:
         width: magnet width [default 10e-3]
         depth: magnet depth [default 20e-3]
         height: magnet height [default 30e-3]
         J: remnant magnetisation in Tesla [default 1.0]

    Optional Arguments:
         centre: magnet centre (tuple or Point3) [default Point3(.0, .0, .0)]
         theta: Angle between magnetisation and x-axis [default 90.0 degrees]
         phi: Angle between magnetisation and z-axis [default 0.0 degrees]  

    """
    mag_type = 'Prism'
    _instances = []
    counter = 0

    def __init__(self,
                 width=10e-3, depth=20e-3, height=30e-3,   # magnet dimensions
                 Jr=1.0,  # local magnetisation direction
                 **kwargs):

        super().__init__(width, depth, height, Jr, **kwargs)

        self.theta = kwargs.pop('theta', 90)
        self.theta_rad = _np.deg2rad(self.theta)
        self.phi = kwargs.pop('phi', 0)
        self.phi_rad = _np.deg2rad(self.phi)

        self.Jx = _np.around(Jr*_np.cos(self.theta_rad)*_np.sin(self.phi_rad),
                             decimals=6)
        self.Jy = _np.around(Jr*_np.sin(self.theta_rad)*_np.sin(self.phi_rad),
                             decimals=6)
        self.Jz = _np.around(Jr*_np.cos(self.phi_rad),
                             decimals=6)
        self.tol = 1e-4  # sufficient for 0.01 degree accuracy

    def __str__(self):
        str = f'{self.__class__.mag_type}\n' + \
              f'J: {self.get_Jr()} (T)\n' + \
              f'Size: {self.size() * 2} (m)\n' + \
              f'Center {self.center()} (m)\n'
        return str

    def __repr__(self):
        str = f'{self.__class__.mag_type}\n' + \
              f'J: {self.get_Jr()} (T)\n' + \
              f'Size: {self.size() * 2} (m)\n' + \
              f'Center {self.center()} (m)\n'
        return str

    def size(self):
        return _np.array([self.a, self.b, self.c])

    def get_Jr(self):
        return _np.array([self.Jx, self.Jy, self.Jz])

    @staticmethod
    def _F1(a, b, c, x, y, z):
        """Helper Function F1 for 3D prismatic magnets

        Args:
            x ([type]): [description]
            y ([type]): [description]
            z ([type]): [description]

        Returns:
            [array]:
        """

        try:
            data = _np.arctan(
                ((y + b) * (z + c)) /
                ((x + a) * _np.sqrt(_np.power((x + a), 2) +
                                    _np.power((y + b), 2) +
                                    _np.power((z + c), 2)))
            )
        except ValueError:
            data = _np.NaN
        return data

    @staticmethod
    def _F2(a, b, c, x, y, z):
        """Helper Function F2 for 3D prismatic magnets

        Args:
            x ([type]): [description]
            y ([type]): [description]
            z ([type]): [description]

        Returns:
            [array]: [description]
        """

        try:
            xa_sq = _np.power((x + a), 2)
            yb_sq = _np.power((y + b), 2)
            zc_sq = _np.power((z + c), 2)
            znc_sq = _np.power((z - c), 2)

            data = (_np.sqrt(xa_sq + yb_sq + znc_sq) + c - z) / (
                _np.sqrt(xa_sq + yb_sq + zc_sq) - c - z)
        except ValueError:
            data = _np.NaN
        return data

    def _calcBx_prism_x(self, a, b, c, Jr, x, y, z):
        """x component of magnetic field for prism magnet
        magnetised in x

        Args:
            a ([type]): [description]
            b ([type]): [description]
            c ([type]): [description]
            Jr ([type]): [description]
            x ([type]): [description]
            y ([type]): [description]
            z ([type]): [description]

        Returns:
            [array]: [description]
        """

        try:
            data = -(Jr/(4 * _np.pi)) * (
                self._F1(a, b, c, -x, y, z) + self._F1(a, b, c, -x, y, -z)
                + self._F1(a, b, c, -x, -y, z) + self._F1(a, b, c, -x, -y, -z)
                + self._F1(a, b, c, x, y, z) + self._F1(a, b, c, x, y, -z)
                + self._F1(a, b, c, x, -y, z) + self._F1(a, b, c, x, -y, -z)
            )
        except ValueError:
            data = _np.NaN
        return data

    def _calcBy_prism_x(self, a, b, c, Jr, x, y, z):
        """y component of magnetic field for prism magnet
        magnetised in x

        Args:
            a ([type]): [description]
            b ([type]): [description]
            c ([type]): [description]
            Jr ([type]): [description]
            x ([type]): [description]
            y ([type]): [description]
            z ([type]): [description]

        Returns:
            [array]: [description]
        """
        # a = self.a
        # b = self.b
        # c = self.c
        # Jr = self.Jr

        try:
            data = _np.log(self._F2(a, b, c, -x, -y, z) *
                           self._F2(a, b, c, x, y, z) /
                           (self._F2(a, b, c, -x, y, z) *
                            self._F2(a, b, c, x, -y, z)))
            data *= (Jr / (4 * _np.pi))
        except ValueError:
            data = _np.NaN
        return data

    def _calcBz_prism_x(self, a, b, c, Jr, x, y, z):
        """z component of magnetic field for prism magnet
        magnetised in x

        Args:
            x ([type]): [description]
            y ([type]): [description]
            z ([type]): [description]

        Returns:
            [array]: [description]
        """
        try:
            data = _np.log(self._F2(a, c, b, -x, -z, y) *
                           self._F2(a, c, b, x, z, y) /
                           (self._F2(a, c, b, -x, z, y) *
                            self._F2(a, c, b, x, -z, y)))
            data *= (Jr/(4 * _np.pi))
        except ValueError:
            data = _np.NaN
        return data

    def _calcB_prism_x(self, x, y, z):
        """Magnetic field vector due to magnet magnetised in x

        Args:
            a ([type]): [description]
            b ([type]): [description]
            c ([type]): [description]
            Jr ([type]): [description]
            x ([type]): [description]
            y ([type]): [description]
            z ([type]): [description]

        Returns:
            [tuple]: Bx, By, Bz
        """
        a = self.a
        b = self.b
        c = self.c
        Jr = self.Jr

        Bx = self._calcBx_prism_x(a, b, c, Jr, x, y, z)
        By = self._calcBy_prism_x(a, b, c, Jr, x, y, z)
        Bz = self._calcBz_prism_x(a, b, c, Jr, x, y, z)
        return Bx, By, Bz

    def _calcB_prism_z(self, x, y, z):
        """Magnetic field vector due to magnet magnetised in y

        Args:
            a ([type]): [description]
            b ([type]): [description]
            c ([type]): [description]
            Jr ([type]): [description]
            x ([type]): [description]
            y ([type]): [description]
            z ([type]): [description]

        Returns:
            [tuple]: Bx, By, Bz
        """
        a = self.a
        b = self.b
        c = self.c
        Jr = self.Jr

        Bx = self._calcBz_prism_x(-c, b, a, Jr, -z, y, x)
        By = self._calcBy_prism_x(-c, b, a, Jr, -z, y, x)
        Bz = -1*self._calcBx_prism_x(-c, b, a, Jr, -z, y, x)
        return Bx, By, Bz

    def _calcB_prism_y(self, x, y, z):
        """Magnetic field vector due to magnet magnetised in z

        Args:
            a ([type]): [description]
            b ([type]): [description]
            c ([type]): [description]
            Jr ([type]): [description]
            x ([type]): [description]
            y ([type]): [description]
            z ([type]): [description]

        Returns:
            [tuple]: Bx, By, Bz
        """
        a = self.a
        b = self.b
        c = self.c
        Jr = self.Jr

        Bx = self._calcBy_prism_x(-b, a, c, Jr, -y, x, z)
        By = -1*self._calcBx_prism_x(-b, a, c, Jr, -y, x, z)
        Bz = self._calcBz_prism_x(-b, a, c, Jr, -y, x, z)
        return Bx, By, Bz


class Cube(Prism):
    """Cube Magnet Class. Cuboid width x depth x height.

   Args:
         width [float]: magnet side length [default 10e-3]
         J: remnant magnetisation in Tesla [default 1.0]

    Optional Arguments:
         centre: magnet centre (tuple or Point3) [default Point3(.0, .0, .0)]
         theta: Angle between magnetisation and x-axis [default 90.0 degrees]
         phi: Angle between magnetisation and z-axis [default 0.0 degrees]  

    """
    mag_type = 'Cube'

    def __init__(self,
                 width=10e-3,   # magnet dimensions
                 Jr=1.0,  # local magnetisation direction
                 **kwargs):

        super().__init__(width=width, depth=width, height=width, Jr=Jr, **kwargs)


class Cylinder(Magnet_3D):
    """Cylinder Magnet Class

    Args:
         radius: radius [default 10e-3]
         length: length [default 10e-3] 
         J: remnant magnetisation in Tesla [default 1.0]

         
    Optional Arguments:
         centre: magnet centre (tuple or Point3) [default Point3(.0, .0, .0)]


    """
    mag_type = 'Cylinder'

    def __init__(self,
                 radius=10e-3, length=10e-3,   # magnet dimensions
                 Jr=1.0,  # local magnetisation direction
                 **kwargs):

        self.radius = radius
        self.length = length
        self.Jr = Jr
        self.Jx = 0
        self.Jy = 0
        self.Jz = Jr

        center = kwargs.pop('center', Point3(0.0, 0.0, 0.0))

        if type(center) is tuple:
            center = Point3(center[0], center[1], center[2])

        self.xc = center.x
        self.yc = center.y
        self.zc = center.z

    def __str__(self):
        str = f'{self.__class__.mag_type}\n' + \
              f'J: {self.Jr} (T)\n' + \
              f'Size: {self.size()} (m)\n' + \
              f'Center {self.center()} (m)\n'
        return str

    def __repr__(self):
        str = f'{self.__class__.mag_type}\n' + \
              f'J: {self.Jr} (T)\n' + \
              f'Size: {self.size() } (m)\n' + \
              f'Center {self.center()} (m)\n'
        return str

    def size(self):
        """Returns magnet dimesions

        Returns:
            size[ndarray]: numpy array [radius, length]
        """
        return _np.array([self.radius, self.length])

    @staticmethod
    @_np.vectorize
    def _cel(kc, p, c, s):
        """ Burlisch's complete elliptic integral
            See NIST Handbook of Mathematical Functions, http://dlmf.nist.gov/19.2
        """
        if kc == 0:
            data = _np.NaN
            return data
        else:
            errtol = 0.000001
            k = _np.abs(kc)
            pp = p
            cc = c
            ss = s
            em = 1.0

            if p > 0:
                pp = _np.sqrt(p)
                ss = s/pp
            else:
                f = kc*kc
                q = 1.0 - f
                g = 1.0 - pp
                f = f - pp
                q = q*(ss - c*pp)
                pp = _np.sqrt(f/g)
                cc = (c - ss)/g
                ss = - q/(g*g*pp) + cc*pp
            f = cc
            cc = cc + ss/pp
            g = k/pp
            ss = 2*(ss + f*g)
            pp = g + pp
            g = em
            em = k + em
            kk = k

            while _np.abs(g - k) > g * errtol:
                k = 2*_np.sqrt(kk)
                kk = k*em
                f = cc
                cc = cc + ss/pp
                g = kk/pp
                ss = 2*(ss + f*g)
                pp = g + pp
                g = em
                em = k + em
            data = (_np.pi / 2.) * (ss + cc * em) / (em * (em + pp))
            return data

    def _calcB_cyl(self, rho, z):
        """ Calculates the magnetic field due to 
            due to a solenoid at any point
            returns Bz,Br
            Omit B0 for B0 = 1 mT
        """
        a = self.radius
        b = self.length / 2
        B0 = self.Jr

        # rho -= self.xc
        # z -= self.zc

        zp = z + b
        zn = z - b

        zp_sq = _np.power(zp, 2)
        zn_sq = _np.power(zn, 2)
        rho_a_sq = _np.power(rho + a, 2)
        nrho_a_sq = _np.power(a - rho, 2)

        alphap = a/_np.sqrt(zp_sq + rho_a_sq)
        alphan = a/_np.sqrt(zn_sq + rho_a_sq)

        betap = zp/_np.sqrt(zp_sq + rho_a_sq)
        betan = zn/_np.sqrt(zn_sq + rho_a_sq)

        gamma = (a - rho) / (a + rho)

        kp = _np.sqrt((zp_sq + nrho_a_sq) / (zp_sq + rho_a_sq))

        kn = _np.sqrt((zn_sq + nrho_a_sq) / (zn_sq + rho_a_sq))

        Br = B0 * (alphap * self._cel(kp, 1, 1, -1)
                   - alphan * self._cel(kn, 1, 1, -1))

        Bz = (B0*a/(a + rho)) * (betap * self._cel(kp, gamma**2, 1, gamma)
                                 - betan * self._cel(kn, gamma**2, 1, gamma))

        return Bz / _np.pi, Br / _np.pi




# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https: // mozilla.org / MPL / 2.0 / .
# Copyright 2021 Peter Dunne
"""Calculate magnetic fields along symmetry axis
of a cylindrical or prismatic magnet
"""
import numpy as _np

__all__ = ['magnetic_field_prism_1D',
           'magnetic_field_cylinder_1D']

def magnetic_field_prism_1D(magnet, z):
    from ._magnet3 import Magnet_3D
    """Magnetic

    Args:
        magnet_object ([type]): [description]
        Jr ([type]): [description]
        z ([type]): [description]

    Returns:
        [type]: [description]
    """
    if issubclass(magnet.__class__, Magnet_3D):
        a = magnet.a
        b = magnet.b
        c = magnet.c
        Jr = magnet.Jr

        if type(z) is not _np.ndarray:
            z_local = _np.asarray(z) - c - magnet.zc
        else:
            z_local = z - c - magnet.zc

        ab = a * b
        a_sq = _np.power(a, 2)
        b_sq = _np.power(b, 2)
        z_sq = _np.power(z_local, 2)
        zc = z_local + 2 * c
        zc_sq = _np.power(zc, 2)

        data = (_np.arctan2(zc * _np.sqrt(a_sq + b_sq + zc_sq), ab) -
                _np.arctan2(z_local * _np.sqrt(a_sq + b_sq + z_sq), ab))
        data *= (Jr / _np.pi)

        return data
    else:
        print(
            f"Error, the magnet should be a 3D magnet not {magnet.__class__}")
        return _np.full_like(z, _np.nan)


def magnetic_field_cylinder_1D(magnet, z):
    from ._magnet3 import Cylinder
    """Magnetic

    Args:
        magnet_object ([type]): [description]
        Jr ([type]): [description]
        z ([type]): [description]

    Returns:
        [type]: [description]
    """
    if issubclass(magnet.__class__, Cylinder):
        L = magnet.length
        R = magnet.radius
        Jr = magnet.Jr
        if type(z) is not _np.ndarray:
            # z_local = z
            z_local = _np.asarray(z) - magnet.length / 2 - magnet.zc
        else:
            # z_local = z
            z_local = z - magnet.length / 2 - magnet.zc

        zL = z_local + L
        R_sq = _np.power(R, 2)
        z_sq = _np.power(z_local, 2)
        zL_sq = _np.power(zL, 2)

        data = ((zL / _np.sqrt(zL_sq + R_sq)) - 
            (z_local / _np.sqrt(z_sq + R_sq)))
        data *= (Jr / 2)
        return data

    else:
        print(
            f"Error, the magnet should be a 3D magnet not {magnet.__class__}")
        return _np.NaN



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https: // mozilla.org / MPL / 2.0 / .
# Copyright 2021 Peter Dunne
"""Example plots
"""
def main():
    print("Line Plot through centre of a cylinder")
    
    import pymagnet as pm
    R = 5e-3
    L = 25e-3
    m_cyl = pm.magnets.Cylinder(radius=R, length=L, Jr=1.2,
                                center=(0.0, 0.0, 0))
    pm.plot.plot_1D_field(m_cyl);

    print("Contour Plot of 2D Magnets")
    pm.reset_magnets()
    cmap = 'viridis' # colormap
    width = 20e-3
    height = 20e-3
    hgap_x = width / 2

    center = (-width / 2 - hgap_x, 0)
    _ = pm.magnets.Rectangle(width=width, height=height,
                            Jr=1.0, center=center, theta=0.0)
    center = (width / 2 + hgap_x, 0)
    _ = pm.magnets.Rectangle(width=width, height=height,
                            Jr=1.0, center=center, theta=90.0)

    x, y = pm.grid2D(2 * width, 2 * height)
    B = pm.B_calc_2D(x, y)


    pm.plot.plot_2D_contour(x, y, B, UL=.6, NL=7, vector_plot=True, cmap=cmap)


if __name__ == '__main__':
    main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https: // mozilla.org / MPL / 2.0 / .
# Copyright 2021 Peter Dunne
# __all__ = ['B_calc_2D']
"""Routines for Two Dimensional Magnet Classes
"""
import numpy as _np

def grid2D(ux, uy, **kwargs):
    """Grid of x,y points.
    Args:
        
        ux ([float]): [x upper limit]
        uy ([float]): [y upper limit]
        lx ([float, optional]): [x lower limit defaults to -ux]
        ly ([optional]): [x lower limit defaults to -ux]
        NP (int, optional): [number of points in each direction]. Defaults to 100.
    """
    NP = kwargs.pop('NP', 100)
    lx = kwargs.pop('lx', -1*ux)
    ly = kwargs.pop('ly', -1*uy)
    NPJ = NP * 1j
    return(_np.mgrid[lx:ux:NPJ, ly:uy:NPJ])


def B_calc_2D(x, y):
    from ._fields import Vector2
    from ._magnet2 import Magnet_2D
    """Function to calculate magnetic field due to any array of points
       It sums the magnetic field B over each component of the magnetisation
       J = mu_0 M
    """
    if (_np.isscalar(x)):
        x = _np.atleast_1d(x)
    if (_np.isscalar(y)):
        y = _np.atleast_1d(y)
    if (_np.ndim(x) == 2):  # planar slice
        B = Vector2(_np.zeros_like(x), _np.zeros_like(x))
    else:
        B = Vector2(_np.zeros(max(x.size, y.size)), _np.zeros(
            max(x.size, y.size)))  # line or single point

    for magnet in Magnet_2D.instances:

        if _np.fabs(magnet.Jx) > magnet.tol:
            B.x += magnet._calcBx_mag_x(x - magnet.xc, y - magnet.yc)
            B.y += magnet._calcBy_mag_x(x - magnet.xc, y - magnet.yc)

        if _np.fabs(magnet.Jy) > magnet.tol:
            B.x += magnet._calcBx_mag_y(x - magnet.xc, y - magnet.yc)
            B.y += magnet._calcBy_mag_y(x - magnet.xc, y - magnet.yc)
    B.calc_norm()
    return B




def _rotvec2(x, y, theta, xc, yc):
    """Rotation of points x,y

    Args:
        x ([ndarray]): array of x points
        y ([ndarray]): array of y points
        theta ([float]): Rotation angle with respect to x-axis
        xc ([float]): translation x co-orinate
        yc ([float]): translation y co-orinate

    Returns:
        x' [ndarray], y' [ndarry]: tuple of array of rotated points
    """
    theta = _np.deg2rad(theta)

    # Rotation matrix, defined for x-y rotation
    R = _np.array([[_np.cos(theta), -_np.sin(theta), 0],
                   [_np.sin(theta), _np.cos(theta), 0], [0, 0, 1]])

    #Define affine transformation for translation
    a = _np.array([[1, 0, xc], [0, 1, yc], [0, 0, 1]])
    c = _np.array([[1, 0, -xc], [0, 1, -yc], [0, 0, 1]])

    # Transformation matrix
    M = _np.linalg.multi_dot([a, R, c])
    b = _np.column_stack((_np.ravel(x),
                          _np.ravel(y),
                          _np.ones_like(_np.ravel(x))))
    rot = _np.dot(M, b.T)

    # pick out the vectors of rotated x- and y-data
    x_rot = rot[0, :]
    y_rot = rot[1, :]
    return _np.reshape(x_rot, x.shape), _np.reshape(y_rot, y.shape)


def _inv_rotvec2(x, y, theta, xc, yc):
    """Inverse rotation of points x,y

    Args:
        x ([ndarray]): array of x points
        y ([ndarray]): array of y points
        theta ([float]): Rotation angle with respect to x-axis
        xc ([float]): translation x co-orinate
        yc ([float]): translation y co-orinate

    Returns:
        x' [ndarray], y' [ndarry]: tuple of array of rotated points
    """    

    theta = _np.deg2rad(theta)
    # Inverse Rotation matrix, defined for x-y rotation
    R = _np.array([[_np.cos(theta), _np.sin(theta), 0],
                   [-_np.sin(theta), _np.cos(theta), 0], [0, 0, 1]]).T

    #Define affine transformation for translation
    a = _np.array([[1, 0, xc], [0, 1, yc], [0, 0, 1]])
    c = _np.array([[1, 0, -xc], [0, 1, -yc], [0, 0, 1]])

    # Transformation matrix
    M = _np.linalg.multi_dot([a, R, c])
    b = _np.column_stack((_np.ravel(x),
                          _np.ravel(y),
                          _np.ones_like(_np.ravel(x))))
    rot = _np.dot(M, b.T)

    # pick out the vectors of rotated x- and y-data
    x_rot = rot[0, :]
    y_rot = rot[1, :]
    return _np.reshape(x_rot, x.shape), _np.reshape(y_rot, y.shape)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https: // mozilla.org / MPL / 2.0 / .
# Copyright 2021 Peter Dunne
"""Two Dimensional Magnet Classes
"""
import numpy as _np
from ._magnet import Magnet
from ._fields import Point2
# from typing import List, Any, Union

__all__ = ['Magnet_2D', 'Rectangle', 'Square']

class Magnet_2D(Magnet):    
    """2D Magnet class.

    Args:
        width [float]: magnet width
        height [float]: magnet height
        Jr [float]: Remnant magnetisation

    Kwargs:
        center [Tuple(float, float), or Point2]: center of magnet, defaults to
        Point2(0.0, 0.0)
    """    
    mag_type = 'Magnet_2D'


    def __init__(self, width, height, Jr, **kwargs) -> None:
        super().__init__()
        self.width = width
        self.height = height
        self.a = width/2
        self.b = height/2
        self.Jr = Jr
        

        center = kwargs.pop('center', Point2(0.0, 0.0))

        if type(center) is tuple:
            center = Point2(center[0], center[1])

        self.xc = center.x
        self.yc = center.y

    def center(self):
        """Returns magnet centre

        Returns:
            center[ndarray]: numpy array [xc, yc]
        """        
        return _np.array([self.xc, self.yc])

    def size(self):
        """Returns magnet dimesions

        Returns:
            size[ndarray]: numpy array [width, height]
        """        
        return _np.array([self.width, self.height])


class Rectangle(Magnet_2D):
    """Rectangle Magnet Class

    Args:
        width [float]: magnet width [m] (defaults to 20e-3)
        height [float]: magnet height [m] (defaults to 40e-3)
        Jr [float]: Remnant magnetisation [T] (defaults to 1.0)
         
    Optional Arguments:
         centre: magnet centre (Point2 object) [default Point2(0.0, 0.0)]
         theta: Angle between magnetisation and x-axis [default 90.0 degrees] 

    """
    mag_type = 'Rectangle'

    def __init__(self,
                 width=20e-3, height=40e-3,  # magnet dimensions
                 Jr=1.0,  # local magnetisation
                 **kwargs):
        super().__init__(width, height, Jr, **kwargs)

        self.theta = kwargs.pop('theta', 90)
        self.theta_rad = _np.deg2rad(self.theta)

        self.Jx = _np.around(Jr * _np.cos(self.theta_rad), decimals=6)
        self.Jy = _np.around(Jr * _np.sin(self.theta_rad), decimals=6)
        self.tol = 1e-4  # sufficient for 0.01 degree accuracy

    def __str__(self):
        str = f'{self.__class__.mag_type}\n' + \
              f'J: {self.get_Jr()} (T)\n' + \
              f'Size: {self.size() * 2} (m)\n' + \
              f'Center {self.center()} (m)\n'
        return str

    def __repr__(self):
        str = f'{self.__class__.mag_type}\n' + \
              f'J: {self.get_Jr()} (T)\n' + \
              f'Size: {self.size()} (m)\n' + \
              f'Center {self.center()} (m)\n'
        return str

    def get_Jr(self):
        """Returns Magnetisation components [Jx, Jy]

        Returns:
            J[ndarray]: [Jx, Jy]
        """        
        return _np.array([self.Jx, self.Jy])

    def _calcBx_mag_x(self, x, y):
        """ Bx using 2D Model for rectangular sheets magnetised in x-plane

        Args:
            magnet ([type]): [magnet object]
            x ([type]): [x array]
            y ([type]): [y array]

        Returns:
            Bx[ndarray]
        """
        a = self.a
        b = self.b
        J = self.Jx
        return (J / (2 * _np.pi)) * (
            _np.arctan2((2 * a * (b + y)), (x**2 - a**2 + (y + b)**2))
            + _np.arctan2((2 * a * (b - y)), (x**2 - a**2 + (y - b)**2))
        )

    def _calcBy_mag_x(self, x, y):
        """ By using 2D Model for rectangular sheets magnetised in x-plane

        Args:
            magnet ([type]): [magnet object]
            x ([type]): [x array]
            y ([type]): [y array]

        Returns:
            [array]: [By]
        """
        a = self.a
        b = self.b
        J = self.Jx
        return (-J / (4 * _np.pi)) * (
            _np.log(((x - a)**2 + (y - b)**2) / ((x + a)**2 + (y - b)**2))
            - _np.log(((x - a)**2 + (y + b)**2) / ((x + a)**2 + (y + b)**2))
        )

    def _calcBx_mag_y(self, x, y):
        """ Bx using 2D Model for rectangular sheets magnetised in y-plane

        Args:
            magnet ([type]): [magnet object]
            x ([type]): [x array]
            y ([type]): [y array]

        Returns:
            [array]: [Bx]
        """
        a = self.a
        b = self.b
        J = self.Jy
        return (J / (4 * _np.pi)) * (
            _np.log(((x + a)**2 + (y - b)**2) / ((x + a)**2 + (y + b)**2))
            - _np.log(((x - a)**2 + (y - b)**2) / ((x - a)**2 + (y + b)**2))
        )

    def _calcBy_mag_y(self, x:float, y: float) -> float:
        """ By using 2D Model for rectangular sheets magnetised in y-plane

        Args:
            magnet ([type]): [magnet object]
            x ([type]): [x array]
            y ([type]): [y array]

        Returns:
            [array]: [By]
        """
        a = self.a
        b = self.b
        J = self.Jy
        return (J / (2 * _np.pi)) * (
            _np.arctan2((2 * b * (x + a)), ((x + a)**2 + y**2 - b**2))
            - _np.arctan2((2 * b * (x - a)), ((x - a)**2 + y**2 - b**2))
        )


class Square(Rectangle):
    """Square Magnet Class

    Args:
        width [float]: magnet side length [m] (defaults to 20e-3)
        Jr [float]: Remnant magnetisation [T] (defaults to 1.0)
         
    Optional Arguments:
         centre: magnet centre (Point2 object) [default Point2(0.0, 0.0)]
         theta: Angle between magnetisation and x-axis [default 90.0 degrees] 

    """
    mag_type = 'Square'

    def __init__(self,
                 width=20e-3,   # magnet dimensions
                 Jr=1.0,  # local magnetisation direction
                 **kwargs):

        super().__init__(width=width, height=width, Jr=Jr, **kwargs)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https: // mozilla.org / MPL / 2.0 / .
# Copyright 2021 Peter Dunne
"""Routines for Three Dimensional Magnet Classes
"""
__all__ = ['B_calc_3D', 'grid3D']

import numpy as _np
from ._fields import Vector3


def grid3D(ux, uy, uz, **kwargs):    
    """Grid of x,y, z points.
    Args:
        
        ux ([float]): [x upper limit]
        uy ([float]): [y upper limit]
        uz ([float]): [z upper limit]
        lx ([float, optional]): [x lower limit defaults to -ux]
        ly ([optional]): [y lower limit defaults to -uy]
        lz ([optional]): [x lower limit defaults to -ux]
        NP (int, optional): [number of points in each direction]. Defaults to 100.

    Returns:
        grid: numpy array
    """
    NP = kwargs.pop('NP', None)

    if NP is None:
        NPx = kwargs.pop('NPx', 100)
        NPy = kwargs.pop('NPy', 100)
        NPz = kwargs.pop('NPz', 100)
    else:
        NPx = NP
        NPy = NP
        NPz = NP

    lx = kwargs.pop('lx', -1*ux)
    ly = kwargs.pop('ly', -1*uy)
    lz = kwargs.pop('lz', -1 * uz)
    return(_np.mgrid[lx:ux:NPx * 1j, ly:uy:NPy * 1j, lz:uz:NPz * 1j])


def B_calc_3D(x, y, z):
    from ._magnet3 import Magnet_3D, Prism, Cube, Cylinder
    """Function to calculate magnetic field due to any array of points
       It sums the magnetic field B over each component of the magnetisation
       J = mu_0 M
    """
    if (_np.isscalar(x)):
        x = _np.atleast_1d(x)
    if (_np.isscalar(y)):
        y = _np.atleast_1d(y)
    if (_np.isscalar(z)):
        z = _np.atleast_1d(z)

    if (_np.ndim(x) == 3):  # volume meshgrid
        B = Vector3(_np.zeros_like(x), _np.zeros_like(x), _np.zeros_like(x))

    elif (_np.ndim(x) == 2):  # planar slice
        B = Vector3(_np.zeros_like(x), _np.zeros_like(x), _np.zeros_like(x))

    elif (_np.ndim(y) == 2):  # planar slice
        B = Vector3(_np.zeros_like(y), _np.zeros_like(y), _np.zeros_like(y))

    else:
        B = Vector3(_np.zeros(max(x.size, y.size, z.size)),
                    _np.zeros(max(x.size, y.size, z.size)),
                    _np.zeros(max(x.size, y.size, z.size)))  # line or single point

    for magnet in Magnet_3D.instances:
        if issubclass(magnet.__class__, Prism):
        
            if _np.fabs(magnet.Jx) > magnet.tol:
                # print('Calc Jx')
                Bx, By, Bz = magnet._calcB_prism_x(x - magnet.xc,
                                        y - magnet.yc,
                                        z - magnet.zc)
                B.x += Bx
                B.y += By
                B.z += Bz

            if _np.fabs(magnet.Jy) > magnet.tol:
                # print('Calc Jy')
                Bx, By, Bz = magnet._calcB_prism_y(x - magnet.xc,
                                        y - magnet.yc,
                                        z - magnet.zc)
                B.x += Bx
                B.y += By
                B.z += Bz

            if _np.fabs(magnet.Jz) > magnet.tol:
                # print('Calc Jz')
                Bx, By, Bz = magnet._calcB_prism_z(x - magnet.xc,
                                        y - magnet.yc,
                                        z - magnet.zc)
                B.x += Bx
                B.y += By
                B.z += Bz
            
        elif issubclass(magnet.__class__, Cylinder):
            print("Cylinder not yet supported")

    B.calc_norm()
    return B


def _rotvec3(x, y, z, theta, phi, xc, yc, zc):
    pass


def _inv_rotvec3(x, y, z, theta, phi, xc, yc, zc):
    pass



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https: // mozilla.org / MPL / 2.0 / .
# Copyright 2021 Peter Dunne
"""Plotting routines


"""
import matplotlib.pyplot as _plt
from matplotlib.patches import Rectangle as _Rect
from matplotlib.patches import Arrow as _Arrow
import numpy as _np
from pymagnet import magnets as _mag

def plot_1D_field(magnet, **kwargs):

    NP = kwargs.pop('NP', 101)
    return_data = kwargs.pop('return_data', False)
    
    if issubclass(magnet.__class__, _mag.Cylinder):
        mag_boundary = magnet.length/2
        z = _np.linspace(-2 * magnet.length +  magnet.zc, 2 * magnet.length + magnet.zc, NP)
        Bz = _mag.magnetic_field_cylinder_1D(magnet, z)
        
    elif issubclass(magnet.__class__, _mag.Prism):
        mag_boundary = magnet.height / 2
        z = _np.linspace(-2*magnet.height + magnet.zc,
                         2*magnet.height + magnet.zc, NP)
        Bz = _mag.magnetic_field_prism_1D(magnet, z)

    else:
        print("Error")
        return None

    _, _ = _plt.subplots()
    _plt.xlabel(r'$z$ (mm)')
    _plt.ylabel(r'$B_z$ (mT)')
    _plt.plot(z*1e3, Bz*1e3,  label='Cube')
    _plt.axvline(x=-mag_boundary*1e3 + magnet.zc*1e3, c='blue', ls='--')
    _plt.axvline(x=mag_boundary*1e3 + magnet.zc*1e3, c='red', ls='--')
    _plt.axvline(x=0.0, c='k', ls='-')
    _plt.show()
    if return_data:
        return z,Bz
    else:
        return None


def plot_2D_line(x, Field, **kwargs):
    """Line Plot of field from 2D magnet

    Args:
        x ([array]): [assumes in m]
        Field ([field vector_2D]): [field vector object]
    """
    scale_x = kwargs.pop('scale_x', 1e-3)
    scale_cb = kwargs.pop('scale_cb', 1)


    xlab = kwargs.pop('xlab', f'x (mm)')
    ylab = kwargs.pop('ylab', f'B (T)')
    axis_scale = kwargs.pop('axis_scale', 'equal')

    SAVE = kwargs.pop('save_fig', False)

    _, ax = _plt.subplots()
    _plt.plot(x/scale_x, Field.n/scale_cb, label=r'$|\mathbf{B}|$')
    _plt.plot(x/scale_x, Field.x/scale_cb, label=r'$B_x$')
    _plt.plot(x/scale_x, Field.y/scale_cb, label=r'$B_y$')
    _plt.legend(loc='best')
    _plt.xlabel(xlab)
    _plt.ylabel(ylab)

    _plt.show()

    if SAVE:
        _plt.savefig('line_plot.png', dpi=300)
        # _plt.savefig('contour_plot.pdf', dpi=300)


def plot_2D_contour(x, y, Field, **kwargs):
    """Contour Plot of field

    Args:
        x ([array]): [assumes in µm]
        y ([array]): [assumes in nm]
        Field ([field vector_2D]): [field vector object]
    """
    scale_x = kwargs.pop('scale_x', 1e-3)
    scale_y = kwargs.pop('scale_y', 1e-3)
    scale_cb = kwargs.pop('scale_cb', 1)

    show_magnets = kwargs.pop('show_magnets', True)

    cmap = kwargs.pop('cmap', 'magma')
    xlab = kwargs.pop('xlab', f'x (mm)')
    ylab = kwargs.pop('ylab', f'y (mm)')
    clab = kwargs.pop('clab', f'B (T)')
    axis_scale = kwargs.pop('axis_scale', 'equal')

    field_component = kwargs.pop('field_component', 'n')

    vector_plot = kwargs.pop('vector_plot', False)
    vector_color = kwargs.pop('vector_color','w')
    NQ = kwargs.pop('vector_arrows', 11)

    if field_component == 'x':
        field_chosen = Field.x
    elif field_component == 'y':
        field_chosen = Field.y
    else:
        field_chosen = Field.n

    SAVE = kwargs.pop('save_fig', False)

    finite_field = field_chosen[_np.isfinite(field_chosen)]/scale_cb
    LL = kwargs.pop('LL', 0)
    UL = kwargs.pop('UL', round(finite_field.mean()*2, 1))
    NL = kwargs.pop('NL', 11)

    lev2 = _np.linspace(LL, UL, 256, endpoint=True)

    _, ax = _plt.subplots()
    CS = _plt.contourf(x/scale_x, y/scale_y, field_chosen/scale_cb,
                       levels=lev2, cmap=_plt.get_cmap(cmap), extend='max')

    if NL > 1:
        lev1 = _np.linspace(LL, UL, NL, endpoint=True)
        _ = _plt.contour(x/scale_x, y/scale_y, field_chosen/scale_cb,
                          vmin=LL, vmax=UL, levels=lev1,
                          linewidths=1.0, colors='k')
        CB = _plt.colorbar(CS, ticks=lev1)
    else:
        CB = _plt.colorbar(CS)


    if vector_plot:
        NPx, NPy = x.shape
        if NQ != 0:
            NSx, NSy = NPx//NQ, NPy//NQ
            _plt.quiver(x[::NSx, ::NSy]/scale_x,
                        y[::NSx, ::NSy]/scale_y,
                        Field.x[::NSx, ::NSy]/Field.n[::NSx, ::NSy],
                        Field.y[::NSx, ::NSy]/Field.n[::NSx, ::NSy],
                        color=vector_color,
                        alpha=1)
    
    if show_magnets:
        patch_array =_num_patch_2D()
        for p in patch_array:
        
            sq = _Rect(xy=(p.patch.x/scale_x, p.patch.y/scale_y),
                       width=p.patch.width/scale_x,
                       height=p.patch.height/scale_y,
                       fill=True, facecolor='#F5F5F5', edgecolor ='k', zorder=5)
            ax.add_patch(sq)

            ar = _Arrow(p.arrow.x/scale_x, p.arrow.y/scale_y,
                        p.arrow.dx/scale_x, p.arrow.dy/scale_y,
                        width = p.arrow.width,
                        zorder=6,
                        color='k')
            ax.add_patch(ar)

    CB.ax.get_yaxis().labelpad = 15
    CB.ax.set_ylabel(clab, rotation=270)
    _plt.axis(axis_scale)
    _plt.xlabel(xlab)
    _plt.ylabel(ylab)
    _plt.show()

    if SAVE:
        _plt.savefig('contour_plot.png', dpi=300)
        # _plt.savefig('contour_plot.pdf', dpi=300)


class patch(object):
    """Encodes magnet dimensions for drawing on plots

    Args:
        object ([type]): [description]
    """    
    def __init__(self, x, y, width, height ):
        super().__init__()
        self.x = x
        self.y = y
        self.width = width
        self.height = height

    def __repr__(self) -> str:
        return f"(x: {self.x}, y: {self.y} w:{self.width}, h: {self.height})"

    def __str__(self) -> str:
        return f"(x: {self.x}, y: {self.y} w:{self.width}, h: {self.height})"

class arrow(object):
    """Encodes magnetisation vector for drawing on plots

    Args:
        object ([type]): [description]
    """    
    def __init__(self, x, y, dx, dy, width=3):
        super().__init__()
        self.x = x
        self.y = y
        self.dx = dx
        self.dy = dy
        self.width = width

    def __repr__(self) -> str:
        return f"(x: {self.x}, y: {self.y}, dx: {self.dx}, dy: {self.dy}, w:{self.width})"

    def __str__(self) -> str:
        return f"(x: {self.x}, y: {self.y}, dx: {self.dx}, dy: {self.dy}, w:{self.width})"

class magnet_patch(object):
    """Magnet drawing class

    Args:
        object ([type]): [description]
    """    
    def __init__(self, patch, arrow)    -> None:
        super().__init__()
        self.patch = patch
        self.arrow = arrow

    def __str__(self) -> str:
        return self.patch.__str__() + self.arrow.__str__()

def _num_patch_2D():
    """Generates patches and arrows for drawing 

    Returns:
        [tuple(list, list)]: lists of patch and arrow objects
    """    
    from .magnets._magnet2 import Magnet_2D
    patch_array = []
    
    for magnet in Magnet_2D.instances:
        patch_tmp = patch(x = magnet.center()[0] - magnet.a,
                          y = magnet.center()[1] - magnet.b,
                          width= 2*magnet.a,
                          height= 2*magnet.b)
        Jnorm = magnet.get_Jr()/_np.abs(magnet.Jr)
        offset = _np.multiply(Jnorm, magnet.size())/2

        arrow_tmp = arrow(x = magnet.center()[0] - offset[0],
                          y = magnet.center()[1] - offset[1],
                          dx= 2 * offset[0],
                          dy= 2 * offset[1])

        magnet_patch_tmp = magnet_patch(patch_tmp, arrow_tmp)

        patch_array.append(magnet_patch_tmp)
    return patch_array


def plot_3D_contour(x, y, z, Field, **kwargs):
    """Contour Plot of field

    Args:
        x ([array]): [assumes in µm]
        y ([array]): [assumes in nm]
        Field ([field vector]): [field vector object]
    """
    scale_x = kwargs.pop('scale_x', 1e-3)
    scale_y = kwargs.pop('scale_y', 1e-3)
    scale_z = kwargs.pop('scale_z', 1e-3)
    scale_cb = kwargs.pop('scale_cb', 1)
    axis_scale = kwargs.pop('axis_scale', 'equal')

    patch_array = kwargs.pop('patch_array', _np.array([]))

    cmap = kwargs.pop('cmap', 'magma')
    xlab = kwargs.pop('xlab', f'x (mm)')
    ylab = kwargs.pop('ylab', f'y (mm)')
    zlab = kwargs.pop('zlab', f'z (mm)')
    clab = kwargs.pop('clab', f'B (T)')

    SAVE = kwargs.pop('save_fig', False)

    finite_field = Field.n[_np.isfinite(Field.n)]/scale_cb
    LL = kwargs.pop('LL', 0)
    UL = kwargs.pop('UL', round(finite_field.mean()*2, 1))
    NL = kwargs.pop('NL', 11)

    lev2 = _np.linspace(LL, UL, 256, endpoint=True)


    if _np.ndim(z) < 2:
        orientation = 'xy'
        plot_x = x/scale_x
        plot_y = y/scale_y
        plot_xlab = xlab
        plot_ylab = ylab

    elif _np.ndim(y) < 2:
        orientation = 'xz'
        plot_x = x/scale_x
        plot_y = z/scale_z
        plot_xlab = xlab
        plot_ylab = zlab

    else:
        orientation = 'yz'
        plot_x = y/scale_y
        plot_y = z/scale_z
        plot_xlab = ylab
        plot_ylab = zlab

    print(orientation)
    _, ax = _plt.subplots()
    CS = _plt.contourf(plot_x, plot_y, Field.n/scale_cb,
                       levels=lev2, cmap=_plt.get_cmap(cmap), extend='max')

    if NL > 1:
        lev1 = _np.linspace(LL, UL, NL, endpoint=True)
        _ = _plt.contour(plot_x, plot_y, Field.n/scale_cb,
                         vmin=LL, vmax=UL, levels=lev1,
                         linewidths=1.0, colors='k')
        CB = _plt.colorbar(CS, ticks=lev1)
    else:
        CB = _plt.colorbar(CS)

    if len(patch_array) > 0:
        for ii in patch_array:
            ii /= scale_x

            sq = _Rect(xy=(ii[0], ii[1]), width=ii[2], height=ii[3],
                       fill=True, color='#F5F5F5', zorder=5)
            ax.add_patch(sq)

            ar = _Arrow(ii[0] + ii[2]/2 - ii[4]/2, ii[1] + ii[3]/2 - ii[5]/2,
                        ii[4], ii[5],
                        width=5, zorder=6,
                        color='k')
            ax.add_patch(ar)

    CB.ax.get_yaxis().labelpad = 15
    CB.ax.set_ylabel(clab, rotation=270)
    _plt.xlabel(plot_xlab)
    _plt.ylabel(plot_ylab)
    _plt.axis(axis_scale)

    if SAVE:
        _plt.savefig('contour_plot.png', dpi=300)
        # _plt.savefig('contour_plot.pdf', dpi=300)


def plot_sub_contour_3D(plot_x, plot_y, plot_B, **kwargs):
    """Contour plot of 3D simulation

    Args:
        plot_x ([type]): [description]
        plot_y ([type]): [description]
        plot_B ([type]): [description]
    """

    cmap = kwargs.pop('cmap', 'seismic')
    xlab = kwargs.pop('xlab', f'x (mm)')
    ylab = kwargs.pop('ylab', f'y (mm)')
    clab = kwargs.pop('clab', f'B (T)')

    axis_scale = kwargs.pop('axis_scale', 'equal')

    SAVE = kwargs.pop('save_fig', False)

    LL = kwargs.pop('LL', -0.5)
    UL = kwargs.pop('UL', 0.5)
    NL = kwargs.pop('NL', 11)

    lev2 = _np.linspace(LL, UL, 256, endpoint=True)
    _, ax = _plt.subplots()
    CS = _plt.contourf(plot_x, plot_y, plot_B,
                      levels=lev2, cmap=_plt.get_cmap(cmap), extend='both')

    if NL > 1:
        lev1 = _np.linspace(LL, UL, NL, endpoint=True)
        _ = _plt.contour(plot_x, plot_y, plot_B,
                        vmin=LL, vmax=UL, levels=lev1,
                        linewidths=1.0, colors='k')
        CB = _plt.colorbar(CS, ticks=lev1)
    else:
        CB = _plt.colorbar(CS)

    CB.ax.get_yaxis().labelpad = 15
    CB.ax.set_ylabel(clab, rotation=270)
    _plt.xlabel(xlab)
    _plt.ylabel(ylab)
    _plt.axis('equal')
    _plt.show()


def line_plot_cylinder(magnet, **kwargs):
    """Calculates and plots the magnetic field along the central axis
    of a cylinder

    This is an example helper function.

    Args:
        magnet ([magnet object]):
    """
    rho = _np.linspace(-2*magnet.radius, 2*magnet.radius, 51)
    z = _np.array([magnet.length * 1.1 / 2])

    Bz, Br = magnet._calcB_cyl(rho, z)
    _, _ = _plt.subplots()
    _plt.plot(rho*1e3, Bz, label='Bz')
    _plt.plot(rho*1e3, Br, label='Br')
    _plt.legend(loc='best')
    _plt.show()

def contour_plot_cylinder(magnet, **kwargs):
    """Calculates and plots the magnetic field 
    of a cylinder

    This is an example helper function.


    Args:
        magnet ([type]): [description]
    """    
    NP = 101
    NPJ = NP*1j
    rho, z = _np.mgrid[-3 * magnet.radius:3 *
                       magnet.radius:NPJ, -magnet.length:magnet.length:NPJ]
    Bz, Br = magnet._calcB_cyl(rho, z)
    Bn = _np.sqrt(Bz**2 + Br**2)

    xlab = f'r (mm)'
    ylab = 'z (mm)'

    # plot_B = Bn
    clab = r'$|B|$ (T)'
    cmap = 'viridis'
    plot_sub_contour_3D(
        rho*1e3, z*1e3, Bn, xlab=xlab, ylab=ylab, clab=clab, cmap=cmap, LL=0, UL=1.0)


def param_test_2D(width, height):
    """Example plots while varying the size of a permanent magnet

    This is an example helper function.

    Args:
        width ([type]): [description]
        height ([type]): [description]
    """    
    x = _np.linspace(-2*width, 2*width, 100)
    y = 1e-3 + height

    B = _mag._routines2.B_calc_2D(x, y)
    plot_2D_line(x, B)
    x, y = _mag._routines2.grid2D(1.5*width, height)
    B = _mag._routines2.B_calc_2D(x, y)
    cmap = 'viridis'
    patch_array = _num_patch_2D()
    plot_2D_contour(
        x, y, B, UL=1, NL=11, cmap=cmap)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https: // mozilla.org / MPL / 2.0 / .
# Copyright 2021 Peter Dunne

import numpy as _np

__all__ = ['Point2', 'Point3', 'Vector2', 'Vector3']


class Point2(object):
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __repr__(self) -> str:
        return f"({self.x}, {self.y})"

    def __str__(self) -> str:
        return f"({self.x}, {self.y})"

    def __add__(self, other):
        return Point2(self.x + other.x, self.y + other.y)
    
    def __sub__(self, other):
        return Point2(self.x - other.x, self.y - other.y)

    def __mul__(self, other):
        return Point2(self.x * other.x, self.y * other.y)
    
    def __div__(self, other):
        x = (self.x / other.x) if other.x != 0 else 0
        y = (self.y / other.y) if other.y != 0 else 0
        return Point2(x, y)

    def __lt__(self, other):
        self_mag = (self.x ** 2) + (self.y ** 2)
        other_mag = (other.x ** 2) + (other.y ** 2)
        return self_mag < other_mag

    def __le__(self, other):
        self_mag = (self.x ** 2) + (self.y ** 2)
        other_mag = (other.x ** 2) + (other.y ** 2)
        return self_mag <= other_mag
    
    def __gt__(self, other):
        self_mag = (self.x ** 2) + (self.y ** 2)
        other_mag = (other.x ** 2) + (other.y ** 2)
        return self_mag > other_mag

    def __ge__(self, other):
        self_mag = (self.x ** 2) + (self.y ** 2)
        other_mag = (other.x ** 2) + (other.y ** 2)
        return self_mag >= other_mag

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y

    def __ne__(self, other):
        return self.x != other.x or self.y != other.y

    def distance_to(self, point):
        return _np.sqrt(_np.power(point.x - self.x, 2) +
                        _np.power(point.y - self.y, 2))

    def distance_to_origin(self):
        return self._norm()
    
    def _norm(self):
        return _np.linalg.norm([self.x, self.y], axis=0)

    def move(self, point):
        try:
            if type(point) is tuple:
                self.x += point[0]
                self.y += point[1]
            else:
                self + point 
        except ValueError as e:
            print("Invalid input, should be a 2-element tuple or Point2 object")




class Point3(Point2):
    def __init__(self, x, y, z):
        super().__init__(x, y)
        self.z = z

    def __repr__(self) -> str:
        return f"({self.x}, {self.y}, {self.z})"

    def __str__(self) -> str:
        return f"({self.x}, {self.y}, {self.z})"
    
    def __add__(self, other):
        return Point3(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other):
        return Point3(self.x - other.x, self.y - other.y, self.z - other.z)

    def __mul__(self, other):
        return Point3(self.x * other.x, self.y * other.y, self.z * other.z)

    def __div__(self, other):
        x = (self.x / other.x) if other.x != 0 else 0
        y = (self.y / other.y) if other.y != 0 else 0
        z = (self.z / other.z) if other.z != 0 else 0
        return Point3(x, y, z)

    def __lt__(self, other):
        self_mag = (self.x ** 2) + (self.y ** 2)
        other_mag = (other.x ** 2) + (other.y ** 2)
        return self_mag < other_mag

    def __le__(self, other):
        self_mag = (self.x ** 2) + (self.y ** 2)
        other_mag = (other.x ** 2) + (other.y ** 2)
        return self_mag <= other_mag

    def __gt__(self, other):
        self_mag = (self.x ** 2) + (self.y ** 2)
        other_mag = (other.x ** 2) + (other.y ** 2)
        return self_mag > other_mag

    def __ge__(self, other):
        self_mag = (self.x ** 2) + (self.y ** 2)
        other_mag = (other.x ** 2) + (other.y ** 2)
        return self_mag >= other_mag

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y

    def __ne__(self, other):
        return self.x != other.x or self.y != other.y

    def distance_to(self, point):
        return _np.sqrt(_np.power(point.x - self.x, 2) +
                        _np.power(point.y - self.y, 2) +
                        _np.power(point.z - self.z, 2))
    
    def _norm(self):
        return _np.linalg.norm([self.x, self.y, self.z], axis=0)

    def distance_to_origin(self):
        return self._norm()


class Vector2(object):
    """2D Field vector class consisting of numpy arrays of x and y values

    Args:
        x:
        y:
    
    Methods:
        calc_norm: calculates the magnitude of the fields at every point and
                    stores in self.n
    """

    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.n = _np.zeros_like(x)

    def calc_norm(self):
        self.n = _np.linalg.norm([self.x, self.y], axis=0)
    
    # def __repr__(self) -> str:
    #     return f"({self.x}, {self.y}, {self.n})"
        
    def __str__(self) -> str:
        return f"({self.x}, {self.y}, {self.n})"


class Vector3(Vector2):
    """3D Field vector class consisting of numpy arrays of x and y values

    Args:
        x:
        y:
        z:
    
    Methods:
        calc_norm: calculates the magnitude of the fields at every point and
                    stores in self.n
    """

    def __init__(self, x, y, z):
        super().__init__(x, y)
        self.z = z

    def calc_norm(self):
        self.n = _np.linalg.norm([self.x, self.y, self.z], axis=0)
    
    # def __repr__(self) -> str:
    #     return f"({self.x}, {self.y}, {self.z}, {self.n})"

    def __str__(self) -> str:
        return f"({self.x}, {self.y}, {self.z}, {self.n})"











       








# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https: // mozilla.org / MPL / 2.0 / .
# Copyright 2021 Peter Dunne
"""Fit data for magnetic field models



Todo:
    * Reimplement all 
"""
pass

# def fit_cube(z, Jr, dz):
#     xDim = 10/2000
#     yDim = 10/2000
#     zDim = 10/2000
#     return(mag.Bzz1d(xDim, yDim, zDim, Jr, z-dz))
#     # return(mag.solBz(xDim,yDim,yDim + y-dy,n,I))


# def fit_cyl(y, Jr, dy):
#     R = 14/2000
#     HH = 10/2000
#     n = 1
#     I = Jr/(u0*n)
#     # return(mag.Byy2d(xDim, yDim, Jr, 0, yDim + y-dy))
#     return(mag.solBz(R, HH, HH+y-dy, n, I))



"""A setuptools based setup module for pymagnet

"""

from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')


setup(
    name='pymagnet',
    version='0.1.0',
    description='User Friendly Magnetic Field Calculations', 
    long_description=long_description,  
    long_description_content_type='text/markdown',
    url='https://github.com/pdunne/pymagnet',
    author='Peter Dunne', 
    author_email='peter.dunne@applied-magnetism.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
    ],

    keywords='magnetism, magnetic fields',
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    python_requires='>=3.6, <4',
    install_requires=['numpy', 'matplotlib'],  # Optional


    # If there are data files included in your packages that need to be
    # installed, specify them here.
    # FIXME:
    # package_data={  # Optional
    #     'sample': ['package_data.dat'],
    # },
    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/distutils/setupscript.html#installing-additional-files
    #
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],  # Optional



    # FIXME:
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/pdunne/pymagnet/issues',
        # 'Funding': 'https://donate.pypi.org',
        # 'Say Thanks!': 'http://saythanks.io/to/example',
        'Source': 'https://github.com/pdunne/pymagnet/',
    },
)
